/**************************************************************

   Program Title:  IMU64X0 Data Logger
   
   Program Name:  IMU64X0_MEMs_Functions.c	  Version: 1.00
   
   Author:  SS	  				  Date:  Feb 22, 2013
   
   Compiled with AVR Studio 4.19
    
   Processor support:  ATmega644, ATmega644P, ATmega1284P
	
   This program will not run properly if DIV8 fuse bit is set (programmed)
   and JTAG pins are enabled - this is the chip default.
		
   The Wasp Monitor is a set of functions that configure Wasp AVR internal
   peripherals for interrupt driven operation.  Custom applications can easily
   be built on top to these basic functions.  Included is a complete data 
   acquistion controller with user preferences loaded into EEPROM.  The Monitor
   communicates with the user via an RS-232 port using the CDP.   The AVR is 
   a smart powerful 8 bit RISC processor that can be easily programmed in C
   with on chip TWI, SPI and analog and digital IO.  Functions are included 
   to read/write data blocks to the on board Serial Flash.  On board sensors
   (if included) include a dual axis 2G accelerometer, temperature and light
   sensor and battery voltage monitor.

   Copyright 2004-2013, SOC Robotics, Inc.  All rights reserved.
   
****************************************************************/
/*
 * Copyright 2004-2013, SOC Robotics, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. Redistribution is only allowed in combination with SOC Machines, 
 *    Inc. hardware products.
 *
 * THIS SOFTWARE IS PROVIDED BY SOC ROBOTICS, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SOC ROBOTICS, 
 * INC. OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
*/
// Include header files for GCC/IAR
#include "avr_compiler.h"
#include <avr/pgmspace.h>

#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
//#include <eeprom.h>
#include <stdio.h>

#include "IMU64X0_Defines.h"
#include "IMU64X0_Type.h"
#include "IMU64X0_Uart.h"
#include "IMU64X0_Extern.h" 
#include "ff.h"
#include "diskio.h"


//#pragma GCC diagnostic warning 

// ***********************************************************************
// ***********************************************************************
//     Start of Code
// ***********************************************************************
// ***********************************************************************
/*
This code was borrowed from 

http://interactive-matter.eu/2009/12/filtering-sensor-data-with-a-kalman-filter/

Implementing a Single Dimension Kalman Filter
The accelerometer puts out three dimensional acceleration data. But if 
you just want to get clean acceleration data from it all formulas simply deal 
with one dimension. So it was easier to use three different one dimensional 
Kalman filters. It all ended in a small set of formulas:

x = x
p = p + q;

k = p / (p + r);
x = x + k * (measurement  x);
p = (1  k) * p;

The first two formulas represent the prediction of the Kalman Filter. And since 
there is no information about the driving forces it is very simple. The second 
three formulas calculate the measurement update. The variables are x for the 
filtered value,  q for the process noise, r for the sensor noise, p for the 
estimated error and k for the Kalman Gain. The state of the filter is defined 
by the values of these variables.

The filter is applied with each measurement and initialized with the process 
noise q, the sensor noise r, the initial estimated error p and the initial 
value x. The initial values for p is not very important since it is adjusted 
during the process. It must be just high enough to narrow down. The initial 
value for the readout is also not very important, since it is updated during 
the process.

But tweaking the values for the process noise and sensor noise is essential 
to get clear readouts.
*/


typedef struct {
	  double q; //process noise covariance
	  double r; //measurement noise covariance
	  double x; //value
	  double p; //estimation error covariance
	  double k; //kalman gain
} kalman_state;

kalman_state volatile filtered_pressure;

kalman_state kalman_init(double q, double r, double p, double intial_value)
{
	  kalman_state result;
	  result.q = q;
	  result.r = r;
	  result.p = p;
	  result.x = intial_value;
 
	  return result;
}

void kalman_update(kalman_state* state, double measurement)
{
	  //prediction update
	  //omit x = x
	  state->p = state->p + state->q;
 
	  //measurement update
	  state->k = state->p / (state->p + state->r);
	  state->x = state->x + state->k * (measurement - state->x);
	  state->p = (1 - state->k) * state->p;
}


/*****************************************************************

  Start of MEMs sensor functions

*****************************************************************/


/*****************************************************************

  BMP085 Sensor Functions

*****************************************************************/

float Finish_Read_BMP085_Pressurefloat(void)
{	 
    unsigned int BMPpressure;
	long x1,x2;
	long b6,x3,b3,p;
	unsigned long b4,b7;
	float compensatedpressure;
	
	// Get raw pressure and compensate
	BMPpressure = Finish_Read_BMP085_Pressure();
	ibmppressure = BMPpressure;  // used in data logger routine
	
	b6 = b5 - 4000;
	x1 = ((long)bmpb2*((b6*b6)/4096))/2048;
	x2 = ((long)bmpac2*b6)/2048;
	x3 = x1 + x2;
	b3 = ((((long)bmpac1*4+x3)<<0)+2)/4;
	x1 = ((long)bmpac3*b6)/8192;
	x2 = ((long)bmpb1*((b6*b6)/4096))/65536;
	x3 = ((x1+x2)+2)/4;
	b4 = ((long)bmpac4*(unsigned long)(x3+32768))/32768;
	b7 = ((unsigned long)BMPpressure-b3)*((unsigned long)50000>>0);
	if(b7<0x80000000) p = (b7*2)/b4;
	else p = (b7/b4)*2;
	x1 = (p/256)*(p/256);
	x1 = (x1*3038)/65536;
	x2 = (-7357*p)/65536;
	p = p + (x1+x2+3791)/16;
	
	compensatedpressure = (float)p*0.00014504;
	return compensatedpressure;
}

float CalculateBMP085Temperature(unsigned int BMPtemperature)
{
	long x1,x2,temperatureinC;

	x1 = ((unsigned long)(BMPtemperature - (long)bmpac6) * (long)bmpac5)/(long)32768;
	x2 = ((long)bmpmc*2048)/(x1+(long)bmpmd);
	b5 = x1 + x2;
	temperatureinC = (b5 + 8)/16;
	return ((float)temperatureinC)/10.0;
}

float CalculateBMP085pressurefloat(unsigned int BMPtemperature,unsigned int BMPpressure)
{	 
	long x1,x2,temperatureinC;
	long b6,x3,b3,p;
	unsigned long b4,b7;
	float compensatedpressure;

	x1 = ((unsigned long)(BMPtemperature - (long)bmpac6) * (long)bmpac5)/(long)32768;
	x2 = ((long)bmpmc*2048)/(x1+(long)bmpmd);
	b5 = x1 + x2;
	temperatureinC = (b5 + 8)/16;
	
	b6 = b5 - 4000;
	x1 = ((long)bmpb2*((b6*b6)/4096))/2048;
	x2 = ((long)bmpac2*b6)/2048;
	x3 = x1 + x2;
	b3 = ((((long)bmpac1*4+x3)<<0)+2)/4;
	x1 = ((long)bmpac3*b6)/8192;
	x2 = ((long)bmpb1*((b6*b6)/4096))/65536;
	x3 = ((x1+x2)+2)/4;
	b4 = ((long)bmpac4*(unsigned long)(x3+32768))/32768;
	b7 = ((unsigned long)BMPpressure-b3)*((unsigned long)50000>>0);
	if(b7<0x80000000) p = (b7*2)/b4;
	else p = (b7/b4)*2;
	x1 = (p/256)*(p/256);
	x1 = (x1*3038)/65536;
	x2 = (-7357*p)/65536;
	p = p + (x1+x2+3791)/16;
	
	compensatedpressure = (float)p*0.00014504;
	return compensatedpressure;
}

float Read_BMP085_pressurefloat(char mode)
{	 
    unsigned int BMPpressure,BMPtemperature;
	long x1,x2,temperatureinC,b5;
	long b6,x3,b3,p;
	unsigned long b4,b7;
	float compensatedpressure;
	
	BMPpressure = BMPtemperature = 0x00;
/*	bmpac1 = 408; bmpac2 = -72; bmpac3 = -14383; bmpac4 = 32741;
	bmpac5 = 32757; bmpac6 = 23151; bmpb1 = 6190; bmpb2 = 4;
	bmpmb = -32768; bmpmc = -8711; bmpmd = 2868;
*/
	
	// Get raw temperature and compensate
 	BMPtemperature = Read_BMP085_Temperature();
	//printf("bmptemp = %d \n\r",BMPtemperature);
	x1 = ((unsigned long)(BMPtemperature - (long)bmpac6) * (long)bmpac5)/(long)32768;
	x2 = ((long)bmpmc*2048)/(x1+(long)bmpmd);
	b5 = x1 + x2;
	temperatureinC = (b5 + 8)/16;
	//printf("b5=%d\n\r",b5);
	
	// Get raw pressure and compensate
	BMPpressure = Read_BMP085_Pressure();
//	BMPpressure = 23843;
	
	b6 = b5 - 4000;
	//printf("b6 = %d\n\r",b6);
	x1 = ((long)bmpb2*((b6*b6)/4096))/2048;
	//printf("x1 = %d\n\r",x1);
	x2 = ((long)bmpac2*b6)/2048;
	//printf("x2 = %d\n\r",x2);
	x3 = x1 + x2;
	//printf("x3 = %d\n\r",x3);
	b3 = ((((long)bmpac1*4+x3)<<0)+2)/4;
	//printf("b3 = %d\n\r",b3);
	x1 = ((long)bmpac3*b6)/8192;
	//printf("x1 = %d\n\r",x1);
	x2 = ((long)bmpb1*((b6*b6)/4096))/65536;
	//printf("x2 = %d\n\r",x2);
	x3 = ((x1+x2)+2)/4;
	//printf("x3 = %d\n\r",x3);
	b4 = ((long)bmpac4*(unsigned long)(x3+32768))/32768;
	//printf("b4 = %u\n\r",b4);
	b7 = ((unsigned long)BMPpressure-b3)*((unsigned long)50000>>0);
	//printf("b7 = %lu\n\r",b7);
	if(b7<0x80000000) p = (b7*2)/b4;
	else p = (b7/b4)*2;
	//printf("p = %ld\n\r",p);
	x1 = (p/256)*(p/256);
	//printf("x1 = %ld\n\r",x1);
	x1 = (x1*3038)/65536;
	//printf("x1 = %d\n\r",x1);
	x2 = (-7357*p)/65536;
	//printf("x2 = %d\n\r",x2);
	p = p + (x1+x2+3791)/16;
	
// 	if(mode==VERBOSE) printf("\rBMP085 temperature raw = %u    \n\r",BMPtemperature);
// 	printf("\rBMP085 temperature (C)= %.2f    \n\r",(float)temperatureinC/10.0);
// 	if(mode==VERBOSE) printf("\rBMP085 pressure raw = %u    \n\r",BMPpressure);
// 	if(mode==VERBOSE) printf("\rBMP085 pressure (p) = %ld    \n\r",p);	
// 	printf("\rBMP085 pressure (psi) = %.4f    \n\r",(float)p*0.00014504);	
	compensatedpressure = (float)p*0.00014504;
	return compensatedpressure;
}


/*****************************************************************

  ADXL345 Sensor Functions

*****************************************************************/

// Set ADXL345 chip select state - PB1
void ADXL345ChipSelect(unsigned char mode)
{
     shortdelay(2);
	 if(mode==LOW) PORTB &= 0xFD; 
	 else PORTB |= 0x02;  
     shortdelay(2);
}

//  Set ADXL345 data format mode
void SPI_ADXL345_DataFormat(void)
{	 
	 //  Set PB1 - Flash CS
	 ADXL345ChipSelect(LOW);

	 // Send read status command
 	 SPI_MasterTransmit(DATAFORMATADDRESS);
 	 SPI_MasterTransmit(0x00);
	 
	 // Set PB1 high - deselect Flash chip
	 ADXL345ChipSelect(HIGH);
}

//  Write command to ADXL345
void SPI_ADXL345_Write(unsigned char commandadd,unsigned char data)
{
	 //  Set PB1 - Flash CS
	 ADXL345ChipSelect(LOW); 

	 // Send read status command
 	 SPI_MasterTransmit(commandadd);
 	 SPI_MasterTransmit(data);
	 
	 // Set PB1 high - deselect Flash chip
	 ADXL345ChipSelect(HIGH);
}

void ADXL345_SetRange(char inchar)
{
 	 if(inchar=='2') SPI_ADXL345_Write(DATAFORMATADDRESS,0x08); 
	 else if(inchar=='4') SPI_ADXL345_Write(DATAFORMATADDRESS,0x09);
	 else if(inchar=='8') SPI_ADXL345_Write(DATAFORMATADDRESS,0x0A);
	 else if(inchar=='6') SPI_ADXL345_Write(DATAFORMATADDRESS,0x0B);
}

// Read acceleration on an axis
int SPI_ADXL345_ReadAxis(unsigned char axisselect)
{
	 unsigned char lowbyte=0,highbyte=0;
	 int axisvalue,axisvalue2;
	 
	 //  Set PB1 - Flash CS
	 ADXL345ChipSelect(LOW);

	 if(axisselect==ADXL345XAXIS) {
 	     SPI_MasterTransmit((XAXISDATA0ADDRESS|0xC0));
	     lowbyte = SPI_MasterReceive();
	     highbyte = SPI_MasterReceive();
		 }
	 else if(axisselect==ADXL345YAXIS) {
 	     SPI_MasterTransmit((YAXISDATA0ADDRESS|0xC0));
	     lowbyte = SPI_MasterReceive();
	     highbyte = SPI_MasterReceive();
		 }
	 else if(axisselect==ADXL345ZAXIS) {
 	     SPI_MasterTransmit((ZAXISDATA0ADDRESS|0xC0));
	     lowbyte = SPI_MasterReceive();
	     highbyte = SPI_MasterReceive();
		 }

	 //printf("low=%x high=%x\n\r",lowbyte,highbyte);
	 axisvalue = 0;
	 axisvalue = lowbyte;
	 axisvalue2 = 0;
	 axisvalue2 = highbyte;
	 axisvalue2 = axisvalue2<<8;
	 axisvalue += axisvalue2;
	 
	 // Set PB1 high - deselect Flash chip
	 ADXL345ChipSelect(HIGH);
	 
	 return axisvalue;
}

//  Get ADXL345 device id
unsigned char SPI_ADXL345_ID(void)
{
 	 unsigned char deviceid;
	 
	 //  Set PB1 - Flash CS
	 ADXL345ChipSelect(LOW);

	 // Send read status command
 	 SPI_MasterTransmit((DEVICEIDADDRESS|0x80));

	 deviceid = SPI_MasterReceive();
	 //printf("device id 1 = %x\n\r",deviceid);
	 
	 // Set PB1 high - deselect Flash chip
	 ADXL345ChipSelect(HIGH);
	 
	 return deviceid;
}

// Display ADXL345 data continuously
void DisplayADXL345Data(void)
{
    int xaxisint,yaxisint,zaxisint,delay;
	float xgrav,ygrav,zgrav,xmin,xmax,ymin,ymax,zmin,zmax;
	float roll,pitch;
	
//	Init_SPI_Master();
	
	//printf("Display gravity - [s-update speed(ms),g-range,e-exit]\n\r");
	//printf("Range is %c\n\r",gmode);
	xmin = ymin = zmin = 16.2;
	xmax = ymax = zmax = -16.2;	
	delay = 1;
		xaxisint = SPI_ADXL345_ReadAxis(ADXL345XAXIS);
		yaxisint = SPI_ADXL345_ReadAxis(ADXL345YAXIS);
		zaxisint = SPI_ADXL345_ReadAxis(ADXL345ZAXIS);
		
		xgrav = (float)xaxisint * 0.0039;	
		ygrav = (float)yaxisint * 0.0039;
		zgrav = (float)zaxisint * 0.0039;
		
		if(xmin>xgrav) xmin = xgrav;
		if(ymin>ygrav) ymin = ygrav;
		if(zmin>zgrav) zmin = zgrav;
		if(xmax<xgrav) xmax = xgrav;
		if(ymax<ygrav) ymax = ygrav;
		if(zmax<zgrav) zmax = zgrav;
		
		//printf("xaxis = %4d ",xaxisint);
		printf("%# 1.3f ",xgrav);
		//printf("yaxis = %4d ",yaxisint);
		printf("%# 1.3f ",ygrav);
		//printf("zaxis = %4d ",zaxisint);
		printf("%# 1.3f  ",zgrav);
		if(0) {
		    printf("%# 1.3f ",xmin);
		    printf("%# 1.3f ",ymin);
		    printf("%# 1.3f ",zmin);
		    printf("%# 1.3f ",xmax);
		    printf("%# 1.3f ",ymax);
		    printf("%# 1.3f ",zmax);	
			}
		if(1) {
		    // Output roll and pitch in degrees
		    roll = asinf(xgrav)*180.0/PI;
			pitch = asinf(ygrav)*180.0/PI;
			printf("%# 2.2f %# 2.2f   ",roll,pitch);
		    }
}


/*****************************************************************

  L3G4200 Gyro Sensor Functions

*****************************************************************/

// Set L3G4200D chip select state - PB2
void L3G4200DChipSelect(unsigned char mode)
{
     shortdelay(2);
	 if(mode==LOW) PORTB &= 0xFB; 
	 else PORTB |= 0x04;  
     shortdelay(2);
}

//  Get L3G4200D register value
void SPI_L3G4200D_PutByte(unsigned char iregister,unsigned char outbyte)
{ 
	 //  Set PB1 - Flash CS
	 L3G4200DChipSelect(LOW);

 	 SPI_MasterTransmit(iregister);
 	 SPI_MasterTransmit(outbyte);
	 
	 // Set PB1 high - deselect Flash chip
	 L3G4200DChipSelect(HIGH);
}

//  Get L3G4200D register value
unsigned char SPI_L3G4200D_GetByte(unsigned char iregister)
{
 	 unsigned char regvalue;
	 
	 //  Set PB1 - Flash CS
	 L3G4200DChipSelect(LOW);

	 // Send who am i command - return value should be D2
 	 SPI_MasterTransmit((iregister|0x80));

	 regvalue = SPI_MasterReceive();
	 //printf("device id = %x\n\r",deviceid);
	 
	 // Set PB1 high - deselect Flash chip
	 L3G4200DChipSelect(HIGH);
	 
	 return regvalue;
}

//  Get L3G4200D register value
int SPI_L3G4200D_ReadAxis(unsigned char axisselect)
{
	 unsigned char lowbyte=0,highbyte=0;
	 int axisvalue,axisvalue2;
	 
	 //  Set PB1 - Flash CS
	 L3G4200DChipSelect(LOW);
 
	 if(axisselect==L3G4200XAXIS) {
 	     SPI_MasterTransmit((OUT_X_L|0xC0));
	     lowbyte = SPI_MasterReceive();
	     highbyte = SPI_MasterReceive();
		 }
	 else if(axisselect==L3G4200YAXIS) {
 	     SPI_MasterTransmit((OUT_Y_L|0xC0));
	     lowbyte = SPI_MasterReceive();
	     highbyte = SPI_MasterReceive();
		 }
	 else if(axisselect==L3G4200ZAXIS) {
 	     SPI_MasterTransmit((OUT_Z_L|0xC0));
	     lowbyte = SPI_MasterReceive();
	     highbyte = SPI_MasterReceive();
		 }

	 // Set PB1 high - deselect Flash chip
	 L3G4200DChipSelect(HIGH);
	 
	 //printf("low=%x high=%x\n\r",lowbyte,highbyte);
	 axisvalue = 0;
	 axisvalue = lowbyte;
	 axisvalue2 = 0;
	 axisvalue2 = highbyte;
	 axisvalue2 = axisvalue2<<8;
	 axisvalue += axisvalue2;
	 
	 return axisvalue;
}

//  Get L3G4200D device id
unsigned char SPI_L3G4200D_ID(void)
{
 	 unsigned char deviceid;
	 
	 //  Set PB1 - Flash CS
	 L3G4200DChipSelect(LOW);

	 // Send who am i command - return value should be D2
 	 SPI_MasterTransmit((WHO_AM_I|0x80));

	 deviceid = SPI_MasterReceive();
	 //printf("device id = %x\n\r",deviceid);
	 
	 // Set PB1 high - deselect Flash chip
	 L3G4200DChipSelect(HIGH);
	 
	 return deviceid;
}

void Set_L3G4200D_Range(char range)
{
    if(range==L250DPS) SPI_L3G4200D_PutByte(CTRL_REG4,0x00);
	if(range==L500DPS) SPI_L3G4200D_PutByte(CTRL_REG4,0x10);
	if(range==L2000DPS) SPI_L3G4200D_PutByte(CTRL_REG4,0x20);
}

void Set_L3G4200D_Bandwidth(char bandwidth)
{
    if(bandwidth==L100HZ) SPI_L3G4200D_PutByte(CTRL_REG1,0x3f);
	if(bandwidth==L200HZ) SPI_L3G4200D_PutByte(CTRL_REG1,0x7f);
	if(bandwidth==L400HZ) SPI_L3G4200D_PutByte(CTRL_REG1,0xBf);
	if(bandwidth==L800HZ) SPI_L3G4200D_PutByte(CTRL_REG1,0xff);
}

/*****************************************************************

  HMC5883 3-Axis Magnetometer Sensor Functions

*****************************************************************/

int HMC5883_ReadAxis(char axis)
{
	union {
	    unsigned char tempbuf[2];
		int outint;
		} mag;

 	if(axis==HMC5883_XAXIS) {
		mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_X_L);
		mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_X_H);
		return mag.outint;
		}
 	if(axis==HMC5883_YAXIS) {
		mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Y_L);
		mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Y_H);
		return mag.outint;
		}
 	if(axis==HMC5883_ZAXIS) {
		mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Z_L);
		mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Z_H);
		return mag.outint;
		}
    return 0;
}

const char hmcmess1[] PROGMEM = "\n\rHMC5883 registers:\n\r";

void DumpHMC5883Regs(void)
{
    unsigned char i;
	
    putfs(hmcmess1);
	for(i=0; i<13; i++) printf(" reg %d: %x\n\r",i,Read_HMC5883_Register(i));
}

void Set_HMC5883_Range(unsigned char rangeselect)
{
    Write_HMC5883_Register(HMC5883_CONFIG_REGB,rangeselect);
}

/*****************************************************************

  MEMs Group Sensor Functions

*****************************************************************/

const char message9[] PROGMEM = "\n\rCheck IMU6410 Sensor presence\n\r";
const char message9b[] PROGMEM = "\n\rCheck IMU6420 Sensor presence\n\r";
const char message10[] PROGMEM = " ADXL345 found,";
const char message11[] PROGMEM = " ADXL345 not found,";
const char message12[] PROGMEM = " L3G4200D found,";
const char message12b[] PROGMEM = " L3GD20 found,";
const char message13[] PROGMEM = " L3G4200D or L2GD20 not found,";
const char message14[] PROGMEM = " HMC5883 found,";
const char message15[] PROGMEM = " HMC5883 not found,";
const char message16[] PROGMEM = " BMP085 found\n\r";
const char message16b[] PROGMEM = " BMP180 found\n\r";
const char message17[] PROGMEM = " BMP085 not found\n\r";
const char message17b[] PROGMEM = " BMP180 not found\n\r";
const char message18[] PROGMEM = " MS5611 found\n\r";
const char message19[] PROGMEM = " MS5611 not found\n\r";

// Check for 10 Degree of freedom DOF10W
void CheckforMEMSensors(void)
{
 	unsigned char ibyte;
	
#ifdef IMU6410
    putfs(message9);
#else
	putfs(message9b);
#endif

	Release_SPI();
	Init_SPI_Master();

    // Check for ADXL345
	if(SPI_ADXL345_ID()==0xE5) {
	    // Init ADXL345
		SPI_ADXL345_Write(DATAFORMATADDRESS,0x0B);
		SPI_ADXL345_Write(POWERCONTROLADDRESS,0x08);
		SPI_ADXL345_Write(INTENABLEADDRESS,0x80);
	    putfs(message10);
		}
	else putfs(message11);
	
	// Check for L3G4200
	if(SPI_L3G4200D_ID()==0xD3) {
	    putfs(message12);
		//SPI_L3G4200D_PutByte(CTRL_REG1,0x0f);  // 100Hz, wake up
		Set_L3G4200D_Bandwidth(L100HZ);  // 100Hz, wake up
		}
	else if(SPI_L3G4200D_ID()==0xD4) {
		putfs(message12b);
		Set_L3G4200D_Bandwidth(L100HZ);  // 100Hz, wake up
		}
	else putfs(message13);
	
	// Check for HMC5883
	Enumerate_HMC5883();
	if(hmc5883state!=0) {
	    putfs(message14);
		ibyte = Read_HMC5883_Register(HMC5883_CONFIG_REGA);
		Write_HMC5883_Register(HMC5883_CONFIG_REGA,(ibyte&0x7F));
		//DumpHMC5883Regs();
		Delaymsec(7);
		}
	else putfs(message15);
	
	// Check for BMP085
	//SetBit(PORTC,PINC2);
	//Delaymsec(100);
	Enumerate_BMP085();
	if(BMPstate!=0) {
		Load_BMP085_Parameters();		
#ifdef  IMU6410
	     putfs(message16);
#else
	     putfs(message16b);
#endif
		}
     else {
#ifdef  IMU6410
	     putfs(message17);
#else
	     putfs(message17b);
#endif
		}

	// Check for MS5611
	Enumerate_MS5611();
	if(ms5611state!=0) {
		Load_MS5611_Parameters();		
	    putfs(message18);
		}
	else putfs(message19);
}

float TruncateAccel(float accel)
{
    if(accel>1.0) return 1.0;
	else if(accel<-1.0) return -1.0;
	else return accel;
}

//  Qaudrant-specific mapping of atanf function.  Uses the sign
//  of y and x to identify which quadrant the result should be
//  mapped to.  Also correctly handles divide-by-zero.
double atan2(double y, double x)
{
 	 if(fabs(x) < 0.0000001) {
	     if(x<0) return -1.570796;
		 else return 1.570796;
		 }
	 if(x<0) {
	     if(y<0) return (atanf(y/x) - PI);
		 else return (atanf(y/x) + PI);
		 }
	 else return atanf(y/x);
}

// Display magnetometer data and compensate compass
// for roll and pitch, display heading
void DisplayMagnetometer(void)
{
    unsigned char i;
	int magx,magy,magz,ax,ay,az;
	float axf,ayf,azf,x2,y2,z2,sxz,syz;
	float magxf,magyf,magzf,magvector,heading;
	float roll,pitch,rollp,pitchp,correctedmagx,correctedmagy;
	union {
	    unsigned char tempbuf[2];
		int outint;
		} mag;

	
	if(0) for(i=3; i<=8; i++) printf("%2x ",Read_HMC5883_Register(i));
	magx = magy = magz = 0;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_X_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_X_H);
	magx = mag.outint;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Y_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Y_H);
	magy = mag.outint;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Z_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Z_H);
	magz = mag.outint;
	if(0) {
	    printf("%4d ",magx);
	    printf("%4d ",magy);
	    printf("%4d ",magz);
		}
		
	magxf = ((float) magx)*1.3/2048.0;
	magyf = ((float) magy)*1.3/2048.0;
	magzf = ((float) magz)*1.3/2048.0;
	// Calculate total magnetic vector value
	magvector = sqrt((magxf*magxf)+(magyf*magyf)+(magzf*magzf));
	//magzf = - magzf;
	// Calculate heading using magx and magy
	// Correct magnetic readings for roll and pitch
	if(1) {
		// Get acceleration raw data
	    ax = SPI_ADXL345_ReadAxis(ADXL345XAXIS);
		ay = SPI_ADXL345_ReadAxis(ADXL345YAXIS);
		az = SPI_ADXL345_ReadAxis(ADXL345ZAXIS);
	    // Bound acceleration to +-g		
		axf = TruncateAccel((float)ax * 0.0039);
		ayf = TruncateAccel((float)ay * 0.0039);
		azf = TruncateAccel((float)az * 0.0039);
		//printf("%5.3f %5.3f %5.3f ",axf,ayf,azf);
		
		x2 = axf * axf;
		y2 = ayf * ayf;
		z2 = azf * azf;
		sxz = sqrt(x2 + z2);
		syz = sqrt(y2 + z2);
		
		//pitch = atan2(axf,sxz);
	    //roll = -atan2(ayf,syz);
		pitch = atan2(axf,azf);
	    roll = atan2(ayf,azf);
		//printf("\n\rroll=%f pitch=%f\n\r",roll,pitch);
		//pitch = atanf(axf/sxz);
	    //roll = -atanf(ayf/syz);
		//printf("roll=%f pitch=%f ",roll,pitch);
		 //_StackCheck(); 
		
		rollp = roll*180.0/PI;
		pitchp = pitch*180.0/PI;
		printf("r=%#6.1f ",rollp);
		printf("p=%#6.1f ",pitchp);
		//  Apply tilt compensation
		
		magxf=-magxf;
		magyf=-magyf;
		
		correctedmagx = magxf*cosf(pitch);
		correctedmagx += magyf*sinf(roll)*sinf(pitch);
		correctedmagx -= magzf*cos(roll)*sinf(pitch);
		correctedmagy = magyf*cosf(roll) + magzf*sinf(roll);
		//printf("%#6.3f ",magxf);
		//printf("%#6.3f ",magyf);
		magxf = correctedmagx;
		magyf = correctedmagy;
		}

	// Calculate heading
	if(magy>0) heading = 90.0 - atanf(magxf/magyf)*180.0/PI;
	else if(magy<0) heading = 270.0 - atanf(magxf/magyf)*180.0/PI;
	else if((magy=0)&&(magx<0)) heading = 180.0;
	else heading = 0.0;
	printf("%#6.3f ",magxf);
	printf("%#6.3f ",magyf);
	printf("%#6.3f ",magzf);
	printf("%#6.3f ",magvector);
	printf("%#5.1f    ",heading);
}

void DisplayMagnetometerPlusMinus(void)
{
    unsigned char i;
	int magx,magy,magz;
	float magxf,magyf,magzf,magvector;
	union {
	    unsigned char tempbuf[2];
		int outint;
		} mag;

	
	if(0) for(i=3; i<=8; i++) printf("%2x ",Read_HMC5883_Register(i));
	// Apply positive bias
	Write_HMC5883_Register(HMC5883_CONFIG_REGA,HMC8SAMPLE|HMCPOSITIVE|HMC1500HZ);
	Delaymsec(100);  // let reading appear
	magx = magy = magz = 0;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_X_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_X_H);
	magx = mag.outint;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Y_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Y_H);
	magy = mag.outint;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Z_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Z_H);
	magz = mag.outint;
	magxf = ((float) magx)*1.3/2048.0;
	magyf = ((float) magy)*1.3/2048.0;
	magzf = ((float) magz)*1.3/2048.0;
	// Calculate total magnetic vector value
	magvector = sqrt((magxf*magxf)+(magyf*magyf)+(magzf*magzf));

	printf("Positive bias:  %#6.3f ",magxf);
	printf("%#6.3f ",magyf);
	printf("%#6.3f ",magzf);
	printf("%#6.3f \n\r",magvector);
	
	// Apply positive bias
	Write_HMC5883_Register(HMC5883_CONFIG_REGA,HMC8SAMPLE|HMCNEGATIVE|HMC1500HZ);
	Delaymsec(100);  // let reading appear
	magx = magy = magz = 0;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_X_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_X_H);
	magx = mag.outint;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Y_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Y_H);
	magy = mag.outint;
	mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Z_L);
	mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Z_H);
	magz = mag.outint;	
	magxf = ((float) magx)*1.3/2048.0;
	magyf = ((float) magy)*1.3/2048.0;
	magzf = ((float) magz)*1.3/2048.0;
	// Calculate total magnetic vector value
	magvector = sqrt((magxf*magxf)+(magyf*magyf)+(magzf*magzf));
	//magzf = - magzf;

	printf("Negative bias:  %#6.3f ",magxf);
	printf("%#6.3f ",magyf);
	printf("%#6.3f ",magzf);
	printf("%#6.3f",magvector);

	Write_HMC5883_Register(HMC5883_CONFIG_REGA,HMC8SAMPLE|HMCNORMAL|HMC1500HZ);
}

void DisplayGyroL3G4200(void)
{
 	unsigned char i;
	int gyrox,gyroy,gyroz;
	float gyroxf,gyroyf,gyrozf;
	union {
	    unsigned char tempbuf[2];
		int outint;
		} gyro;
		
	for(i=OUT_X_L; i<=OUT_Z_H; i++) printf("%2x ",SPI_L3G4200D_GetByte(i));
	gyrox = gyroy = gyroz = 0;
	gyro.tempbuf[0] = SPI_L3G4200D_GetByte(OUT_X_L);
	gyro.tempbuf[1] = SPI_L3G4200D_GetByte(OUT_X_H);
	gyrox = gyro.outint;
	gyro.tempbuf[0] = SPI_L3G4200D_GetByte(OUT_Y_L);
	gyro.tempbuf[1] = SPI_L3G4200D_GetByte(OUT_Y_H);
	gyroy = gyro.outint;
	gyro.tempbuf[0] = SPI_L3G4200D_GetByte(OUT_Z_L);
	gyro.tempbuf[1] = SPI_L3G4200D_GetByte(OUT_Z_H);
	gyroz = gyro.outint;
	printf("%6d ",gyrox);
	printf("%6d ",gyroy);
	printf("%6d ",gyroz);
	gyroxf = ((float) gyrox)*250.0/32768.0;
	gyroyf = ((float) gyroy)*250.0/32768.0;
	gyrozf = ((float) gyroz)*250.0/32768.0;
	printf("%#6.1f ",gyroxf);
	printf("%#6.1f ",gyroyf);
	printf("%#6.1f       ",gyrozf);
}

void DumpL3G4200Registers(void)
{
 	 unsigned char i;
	 printf("\n\rL3G4200D Registers\n\r");
	 printf(" reg 0x0f: %x\n\r",SPI_L3G4200D_GetByte(0x0f));
	 for(i=0x20; i<=0x38; i++)  printf(" reg %x: %x\n\r",i,SPI_L3G4200D_GetByte(i));
}

void DisplayBMP085Brief(char mode)
{
	long x1,x2,b5,temperatureinC,BMPtemperature;
	float pressure;
	union {
	    unsigned char tempbuf[4];
		float pressure;
		} baro;
	union {
	    unsigned char tempbuf[4];
		float temperature;
		} temp;
	
    // Get raw temperature and compensate
 	BMPtemperature = Read_BMP085_Temperature();
	if(0) printf("\n\rbmptemp=%ld\n\r",BMPtemperature);
//	BMPtemperature = (long)33129;
//	printf("\n\rbmptemp=%u\n\r",BMPtemperature);
	x1 = ((long)(BMPtemperature - (long)bmpac6) * (long)bmpac5)/(long)32768;
	x2 = ((long)bmpmc*2048)/(x1+(long)bmpmd);
	b5 = x1 + x2;
	if(0) {
	    printf("x1=%ld ",x1);
	    printf("x2=%ld ",x2);
	    printf("b5=%ld\n\r",b5);
		}
	temperatureinC = (b5 + 8)/16;
	//printf("temp = %d\n\r",temperatureinC);
	temp.temperature = ((float) temperatureinC)/10.0;
	
	pressure = Read_BMP085_pressurefloat(QUIET);
	baro.pressure = pressure;
	
	if(mode=='b') {
	    putchar(temp.tempbuf[0]);
	    putchar(temp.tempbuf[1]);
	    putchar(temp.tempbuf[2]);
	    putchar(temp.tempbuf[3]);
	    putchar(baro.tempbuf[0]);
	    putchar(baro.tempbuf[1]);
	    putchar(baro.tempbuf[2]);
	    putchar(baro.tempbuf[3]);
	    }
 	else {
	    printf("%#5.2f ",((float)temperatureinC)/10.0);
 	    printf("%#6.4f %#6.1f",pressure,(14.696-pressure)/0.000407);	
		}
}

void printtime(void)
{
    unsigned char instrg[5];
	float floattime;
	
    cli();
		instrg[0] = days;
		instrg[1] = hours;
		instrg[2] = minutes;
		instrg[3] = seconds;
		instrg[4] = rtcclockcnt;
	sei();	
		
	floattime = (float) seconds + ((float) (instrg[4]*4))/1000.0;
	printf("%2d:%2d:%#2.3f ", instrg[1],instrg[2],floattime);
}

//  BMP085 Read Update State Machine
//  The BMP085 takes 5msec to update the temperature ready from 
//  a read request and the pressure takes about 28msec but both
//  parameters are read using the same read register so the state 
//  machine needs to keep track of which parameter is being accessed
//  and not hold up the access of other mems sensors.
//  This version is used by the datalogger routine
void Check_BMP085_Update_Int(void)
{  	 
	 // See which parameter is active
     if(bmp085state==BMPTEMP) { // Update temperature
	     // If temp reading hasn't started - start it
	     if(bmp085tempstate==BMPTEMPIDLE) {
		     bmp085tempstate = BMPTEMPREAD;
			 bmp085countdown = 8;
			 Start_Read_BMP085_Temperature();
			 //printf("Start read temp\n\r");
			 }
	     // Else grab the temperature reading and switch sensors
		 else if(bmp085countdown==0) {
		     ibmptemperature = Finish_Read_BMP085_Temperature();

			 //printf("Read temp = %f\n\r",bmptemperature);

			 // Number of pressure readings to take before reading temp again
		     bmpskipcount = BMPSKIPVALUE;  
			 bmp085state = BMPPRESSURE;
			 bmp085tempstate = BMPTEMPIDLE;
		     }
		 }
	 else {
	     if(bmp085pressurestate==BMPPRESSUREIDLE) {
		     bmp085pressurestate = BMPPRESSUREREAD;
			 bmp085countdown = 28;
			 Start_Read_BMP085_Pressure();
			 //printf("Start read pressure\n\r");
		     }		
		 else {
		     if(bmp085countdown==0) {
				 ibmppressure = Finish_Read_BMP085_Pressure();
			     //printf("Read pressure = %f\n\r",bmppressure);
				 if(bmpskipcount--==0) {
				     bmp085state = BMPTEMP;
			         bmp085pressurestate = BMPPRESSUREIDLE;
					 }
				 else {
				     Start_Read_BMP085_Pressure();
			 		 //printf("Start read pressure\n\r");
				     bmp085countdown = 28;
					 }
			     }
			 }
		 }
}		

//  BMP085 Read Update State Machine
//  The BMP085 takes 5msec to update the temperature ready from 
//  a read request and the pressure takes about 28msec but both
//  parameters are read using the same read register so the state 
//  machine needs to keep track of which parameter is being accessed
//  and not hold up the access of other mems sensors.
void Check_BMP085_Update(void)
{  
     unsigned int BMPtemperature;
     long x1,x2,temperatureinC;
	 
	 // See which parameter is active
     if(bmp085state==BMPTEMP) { // Update temperature
	     // If temp reading hasn't started - start it
	     if(bmp085tempstate==BMPTEMPIDLE) {
		     bmp085tempstate = BMPTEMPREAD;
			 bmp085countdown = 8;
			 Start_Read_BMP085_Temperature();
			 //printf("Start read temp\n\r");
			 }
	     // Else grab the temperature reading and switch sensors
		 else if(bmp085countdown==0) {
		     BMPtemperature = Finish_Read_BMP085_Temperature();
			 ibmptemperature = BMPtemperature;  // used in data logger routine
		     x1 = ((long)(BMPtemperature - (long)bmpac6) * (long)bmpac5)/(long)32768;
		     x2 = ((long)bmpmc*2048)/(x1+(long)bmpmd);
		     b5 = x1 + x2;
		     temperatureinC = (b5 + 8)/16;
		     bmptemperature = ((float)temperatureinC)/10.0;
			 //printf("\n\rb5=%d ",b5);
			 //printf("temp=%f\n\r",bmptemperature);

			 //printf("Read temp = %f\n\r",bmptemperature);

			 // Number of pressure readings to take before reading temp again
		     bmpskipcount = BMPSKIPVALUE;  
			 bmp085state = BMPPRESSURE;
			 bmp085tempstate = BMPTEMPIDLE;
		     }
		 }
	 else {
	     if(bmp085pressurestate==BMPPRESSUREIDLE) {
		     bmp085pressurestate = BMPPRESSUREREAD;
			 bmp085countdown = 28;
			 Start_Read_BMP085_Pressure();
			 //printf("Start read pressure\n\r");
		     }		
		 else {
		     if(bmp085countdown==0) {
				 bmppressure = Finish_Read_BMP085_Pressurefloat();
			     //printf("Read pressure = %f\n\r",bmppressure);
				 if(bmpskipcount--==0) {
				     bmp085state = BMPTEMP;
			         bmp085pressurestate = BMPPRESSUREIDLE;
					 }
				 else {
				     Start_Read_BMP085_Pressure();
			 		 //printf("Start read pressure\n\r");
				     bmp085countdown = 28;
					 }
			     }
			 }
		 }
}		

// Cycle the BMP085 update state machine to load 
// initial temperature and pressure readings
void Init_BMP085_Readings()
{
	bmpskipcount = BMPSKIPVALUE;  
	bmp085state = BMPTEMP;
	bmp085tempstate = BMPTEMPIDLE;
	bmp085pressurestate = BMPPRESSUREIDLE;
	Check_BMP085_Update();
	Delaymsec(9); // wait for bmp085 temp reading
	Check_BMP085_Update();  // get temperature reading
	Check_BMP085_Update();
	Delaymsec(30); // wait for bmp085 pressure reading
	Check_BMP085_Update();  // get pressure reading
}

extern BYTE Buff[];
extern FIL File[];

// Display accelerometer, gyro, magnetometer and pressure
void DisplayDOF10WData(char mode)
{	
	unsigned char overflow,outstrg[100],sensorrange;
	unsigned int s2;
	int gyrox,gyroy,gyroz,xaxisint,yaxisint,zaxisint;
	int magx,magy,magz,i;
	float gyroxf,gyroyf,gyrozf,xgrav,ygrav,zgrav,magxf,magyf,magzf;
	float agyroxf,agyroyf,agyrozf,axgrav,aygrav,azgrav,amagxf,amagyf,amagzf;
	float floattime,gyroscale,magscale,percent;


    union {
	    unsigned char tempbuf[2];
		int outint;
		} accel;
	union {
	    unsigned char tempbuf[2];
		int outint;
		} gyro;
	union {
	    unsigned char tempbuf[2];
		int outint;
		} mag;
	union {
	    unsigned char tempbuf[4];
		float pressure;
		} baro;
	union {
	    unsigned char tempbuf[4];
		float temperature;
		} temp;

	//printf("xa  ya   za    xg    yg   zg   xm   ym   zm   t  psi\n\r");
	if((findthresholdaverage==YES)&&(thresholdloopcnt==0)) {
		agyroxf=agyroyf=agyrozf=axgrav=aygrav=azgrav=amagxf=amagyf=amagzf=0.0;
	    }

	// Check if single pass mode selected and convert to t mode -
	// calling routine will terminate loop
    if(mode=='s') mode = 't';

	cli();
		//instrg[0] = days;
		outstrg[0] = hours;
		outstrg[1] = minutes;
		outstrg[2] = seconds;
		outstrg[3] = rtcclockcnt;
	sei();	
	
    if((mode=='r')||(mode=='t')) { // output time
		floattime = (float) outstrg[2] + ((float) ((int)outstrg[3]*3.90625))/1000.0;
        //printf("%02d:%02d:%02.3f ", outstrg[0],outstrg[1],floattime);
		sprintf(outstrg,"%02d:%02d:%02.3f ", outstrg[0],outstrg[1],floattime);
		strcpy(Buff,outstrg);
		strcpy(time_tracker,outstrg);
	    }

	overflow = 0;

	// Calculate scale factor based on sensor range setting
	sensorrange = SensorRangeByte();
	// If ADXL set to full resolution mode 4mg resolution maintained on all ranges
	//accelscale = CalculateAccelScale(sensorrange);
	gyroscale = CalculateGyroScale(sensorrange);
	magscale = CalculateMagScale(sensorrange);
		
	// Display accelerometer data
	xaxisint = SPI_ADXL345_ReadAxis(ADXL345XAXIS);
	yaxisint = SPI_ADXL345_ReadAxis(ADXL345YAXIS);
	zaxisint = SPI_ADXL345_ReadAxis(ADXL345ZAXIS);
	
	xgrav = (float)xaxisint * 0.0039;	
	ygrav = (float)yaxisint * 0.0039;
	zgrav = (float)zaxisint * 0.0039;
	if(findthresholdaverage==YES) {
	    axgrav += xgrav;
		aygrav += ygrav;
		azgrav += zgrav;
		}
	
	if(first_time_grav == YES) {
		max_values[0] = min_values[0] = xgrav;
		max_values[1] = min_values[1] = ygrav;
		max_values[2] = min_values[2] = zgrav;
	}

	if(mode=='b') {
	    accel.outint = xaxisint;
		outstrg[4] = accel.tempbuf[0];
		outstrg[5] = accel.tempbuf[1];
	    accel.outint = yaxisint;
		outstrg[6] = accel.tempbuf[0];
		outstrg[7] = accel.tempbuf[1];
	    accel.outint = zaxisint;
		outstrg[8] = accel.tempbuf[0];
		outstrg[9] = accel.tempbuf[1];
	    }
	else if(mode=='t') {
		sprintf(outstrg,"%#6.3f %#6.3f %#6.3f ",xgrav,ygrav,zgrav);
		strcat(Buff,outstrg);
		}
	else if(mode=='r') {
	    printf("%4d ",xaxisint);
	    printf("%4d ",yaxisint);
	    printf("%4d  ",zaxisint);
		}
	//if((xgrav >= max_values[0])	|| (xgrav <= min_values[0]) || (ygrav >= max_values[1])
	//|| (ygrav <= min_values[1])	|| (zgrav >= max_values[2])	|| (zgrav <= min_values[2]))
	//{
	// Check min/max's
		if(xgrav >= max_values[0]) {
			max_values[0] = xgrav;
			strcpy(max_time_tracker0, time_tracker);
			}
		if(xgrav <= min_values[0]) {
			min_values[0] = xgrav;
			strcpy(min_time_tracker0, time_tracker);
			}
		if(ygrav >= max_values[1]) {
			max_values[1] = ygrav;
			strcpy(max_time_tracker1, time_tracker);
			}
		if(ygrav <= min_values[1]) {
			min_values[1] = ygrav;
			strcpy(min_time_tracker1, time_tracker);
			}
		if(zgrav >= max_values[2]) {
			max_values[2] = zgrav;
			strcpy(max_time_tracker2, time_tracker);
			}
		if(zgrav <= min_values[2]) {
			min_values[2] = zgrav;
			strcpy(min_time_tracker2, time_tracker);
			}	
	//}

    // Display L3G4200D gyro data
	gyrox = gyroy = gyroz = 0;
	gyrox = SPI_L3G4200D_ReadAxis(L3G4200XAXIS);
	gyroy = SPI_L3G4200D_ReadAxis(L3G4200YAXIS);
	gyroz = SPI_L3G4200D_ReadAxis(L3G4200ZAXIS);
	gyroxf = ((float) gyrox)*250.0*gyroscale/32768.0;
	gyroyf = ((float) gyroy)*250.0*gyroscale/32768.0;
	gyrozf = ((float) gyroz)*250.0*gyroscale/32768.0;
	if(findthresholdaverage==YES) {
	    agyroxf += gyroxf;
	    agyroyf += gyroyf;
	    agyrozf += gyrozf;
		}

	if(first_time_gyro == YES) {
		max_values[3] = min_values[3] = gyroxf;
		max_values[4] = min_values[4] = gyroyf;
		max_values[5] = min_values[5] = gyrozf;
		}

	if(mode=='t') {
		sprintf(outstrg,"%#6.1f %#6.1f %#6.1f ",gyroxf,gyroyf,gyrozf);
		strcat(Buff,outstrg);
		}
	else if(mode=='r') {
	    printf("%4d ",gyrox);
	    printf("%4d ",gyroy);
	    printf("%4d  ",gyroz);
		}
	else if(mode=='b') {
	    gyro.outint = gyrox;
	    outstrg[10] = gyro.tempbuf[0];
	    outstrg[11] = gyro.tempbuf[1];
	    gyro.outint = gyroy;
	    outstrg[12] = gyro.tempbuf[0];
	    outstrg[13] = gyro.tempbuf[1];
	    gyro.outint = gyroz;
	    outstrg[14] = gyro.tempbuf[0];
	    outstrg[15] = gyro.tempbuf[1];
		}

	//if((gyroxf >= max_values[3]) || (gyroxf <= min_values[3]) || (gyroyf >= max_values[4])
	//|| (gyroyf <= min_values[4]) || (gyrozf >= max_values[5]) || (gyrozf <= min_values[5]))
	//{
	// Check min/max's
		if(gyroxf >= max_values[3]) {
			max_values[3] = gyroxf;
			strcpy(max_time_tracker3, time_tracker);
			}
		if(gyroxf <= min_values[3])	{
			min_values[3] = gyroxf;
			strcpy(min_time_tracker3, time_tracker);
			}
		if(gyroyf >= max_values[4]) {
			max_values[4] = gyroyf;
			strcpy(max_time_tracker4, time_tracker);
			}
		if(gyroyf <= min_values[4]) {
			min_values[4] = gyroyf;
			strcpy(min_time_tracker4, time_tracker);
			}
		if(gyrozf >= max_values[5]) {
			max_values[5] = gyrozf;
			strcpy(max_time_tracker5, time_tracker);
			}
		if(gyrozf <= min_values[5]) {
			min_values[5] = gyrozf;
			strcpy(min_time_tracker5, time_tracker);
			}
	//}

    // Display magnetometer data
	outstrg[16] = mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_X_L);
	outstrg[17] = mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_X_H);
	magx = mag.outint;
	outstrg[18] = mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Y_L);
	outstrg[19] = mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Y_H);
	magy = mag.outint;
	outstrg[20] = mag.tempbuf[0] = Read_HMC5883_Register(HMC5883_OUT_Z_L);
	outstrg[21] = mag.tempbuf[1] = Read_HMC5883_Register(HMC5883_OUT_Z_H);
	magz = mag.outint;

	magxf = ((float) magx)*1.3*magscale/2048.0;
	magyf = ((float) magy)*1.3*magscale/2048.0;
	magzf = ((float) magz)*1.3*magscale/2048.0;
	if(findthresholdaverage==YES) {
	    amagxf += magxf;
	    amagyf += magyf;
	    amagzf += magzf;
		}
	
	if(first_time_mag == YES) {
		max_values[6] = min_values[6] = magxf;
		max_values[7] = min_values[7] = magyf;
		max_values[8] = min_values[8] = magzf;
		}

	if(mode=='t') {
		sprintf(outstrg,"%#6.3f %#6.3f %#6.3f ",magxf,magyf,magzf);
		strcat(Buff,outstrg);
		}
	else if(mode=='r') {
	    printf("%4d ",magx);
	    printf("%4d ",magy);
	    printf("%4d  ",magz);
		}

	//if((magxf >= max_values[6]) || (magxf <= min_values[6]) || (magyf >= max_values[7])
	//|| (magyf <= min_values[7]) || (magzf >= max_values[8]) || (magzf <= min_values[8]))
	//{	
		if(magxf >= max_values[6]) {
			max_values[6] = magxf;
			strcpy(max_time_tracker6, time_tracker);
			}
		if(magxf <= min_values[6]) {
			min_values[6] = magxf;
			strcpy(min_time_tracker6, time_tracker);
			}
		if(magyf >= max_values[7]) {
			max_values[7] = magyf;
			strcpy(max_time_tracker7, time_tracker);
			}
		if(magyf <= min_values[7]) {
			min_values[7] = magyf;
			strcpy(min_time_tracker7, time_tracker);
			}
		if(magzf >= max_values[8]) {
			max_values[8] = magzf;
			strcpy(max_time_tracker8, time_tracker);
			}
		if(magzf <= min_values[8]) {
			min_values[8] = magzf;
			strcpy(min_time_tracker8, time_tracker);
			}
	//}

	// Update BMP085 temperature once per second
	// and pressure once every 28msec or longer
	if(bmp085mode==ACTIVE) {
	    // Display barometer data
		Check_BMP085_Update();
		if(mode!='b') {
			sprintf(outstrg,"%#4.1f %#6.4f ",bmptemperature,bmppressure);
			strcat(Buff,outstrg);
			}

	    else {
		    temp.temperature = bmptemperature;
		    baro.pressure = bmppressure;
	        outstrg[22] = temp.tempbuf[0];
    	    outstrg[23] = temp.tempbuf[1];
			outstrg[24] = temp.tempbuf[2];
			outstrg[25] = temp.tempbuf[3];
    	    outstrg[26] = baro.tempbuf[0];
    	    outstrg[27] = baro.tempbuf[1];
    	    outstrg[28] = baro.tempbuf[2];
    	    outstrg[29] = baro.tempbuf[3];
			outstrg[30] = '\r';
	        }
		}
	else outstrg[22] = '\r';
	
	if(mode=='b')  {
	    if(bmp085mode==ACTIVE)  for(i=0; i<30; i++)  putchar(outstrg[i]);
		else for(i=0; i<23; i++)  putchar(outstrg[i]);
		}
	else {
		if(printoutput==YES)  {
			if((thresholdmode==ACTIVE)&&(foundthreshold==YES)) {
			    printf("%s",Buff);
		        putchar('\r');
				}
			else if(thresholdmode==INACTIVE){
			    printf("%s",Buff);
		        putchar('\r');
				}
			}
		}

	if(thresholdmode==ACTIVE) {
		if((findthresholdaverage==YES)&&(thresholdloopcnt++>=10)) {
			axgrav = axgrav/10.0;
			aygrav = aygrav/10.0;
			azgrav = azgrav/10.0;
			agyroxf = agyroxf/10.0;
			agyroyf = agyroyf/10.0;
			agyrozf = agyrozf/10.0;
			amagxf = amagxf/10.0;
			amagyf = amagyf/10.0;
			amagzf = amagzf/10.0;
			findthresholdaverage = NO;	
		    }
    	else if((foundthreshold==NO)&&(thresholdloopcnt++>=10)) {
			percent = thresholdpercent/100.0;
		    if(thresholdcheck[0]=='y') {
			    if(xgrav>(axgrav+(2.00*percent))) foundthreshold = YES;
				if(xgrav<(axgrav-(2.00*percent))) foundthreshold = YES;
				//printf("\n\rxgrav=%f %f %f \n\r",xgrav,axgrav+(2.00*percent),axgrav-(2.00*percent));
				}
		    if(thresholdcheck[1]=='y') {
			    if(ygrav>(azgrav+(2.00*percent))) foundthreshold = YES;
				if(ygrav<(azgrav-(2.00*percent))) foundthreshold = YES;
				}
		    if(thresholdcheck[2]=='y') {
			    if(zgrav>(azgrav+(2.00*percent))) foundthreshold = YES;
				if(zgrav<(azgrav-(2.00*percent))) foundthreshold = YES;
				}
		    if(thresholdcheck[3]=='y') {
			    if(gyroxf>(agyroxf+(250.0*percent))) foundthreshold = YES;
				if(gyroxf<(agyroxf-(250.0*percent))) foundthreshold = YES;
				}
		    if(thresholdcheck[4]=='y') {
			    if(gyroyf>(agyroyf+(250.0*percent))) foundthreshold = YES;
				if(gyroyf<(agyroyf-(250.0*percent))) foundthreshold = YES;
				}
		    if(thresholdcheck[5]=='y') {
			    if(gyrozf>(agyrozf+(250.0*percent))) foundthreshold = YES;
				if(gyrozf<(agyrozf-(250.0*percent))) foundthreshold = YES;
				}
		    if(thresholdcheck[6]=='y') {
			    if(magxf>(amagxf+(1.00*percent))) foundthreshold = YES;
				if(magxf<(amagxf-(1.00*percent))) foundthreshold = YES;
				}
		    if(thresholdcheck[7]=='y') {
			    if(magyf>(amagyf+(1.00*percent))) foundthreshold = YES;
				if(magyf<(amagyf-(1.00*percent))) foundthreshold = YES;
				}
		    if(thresholdcheck[8]=='y') {
			    if(magzf>(amagzf+(1.00*percent))) foundthreshold = YES;
				if(magzf<(amagzf-(1.00*percent))) foundthreshold = YES;
				}
			if(foundthreshold==YES) printf("Threshold found\n\r");
			}
		else if(storageselection==MICROSD) {
			strcat(Buff,"\r\n");
			f_write(&File[0], Buff, strlen(Buff), &s2);
		}
    }
	else if(storageselection==MICROSD) {
		strcat(Buff,"\r\n");
		f_write(&File[0], Buff, strlen(Buff), &s2);
		// To stop file loss due to power failure flush the buff using
		// f_sync call every FSYNCMAX cycles - default is 200
		// f_sycn introduces a 10msec delay
		if(fsynccnt++>FSYNCMAX)  {
		    f_sync(&File[0]);
			fsynccnt = 0;
			}
	    }
}

/*****************************************************************

  MEMs Filter Functions

*****************************************************************/
volatile float GyroBias[3],AccelBias[3];

typedef struct {
  		int xaccel; 
  		int yaccel; 
  		int zaccel; 
  		int xgyro; 
  		int ygyro; 
  		int zgyro; 
  		int xmag;
  		int ymag;
  		int zmag;
  		float xfaccel; 
  		float yfaccel; 
  		float zfaccel; 
  		float xfgyro; 
  		float yfgyro; 
  		float zfgyro; 
  		float xfmag;
  		float yfmag;
  		float zfmag;
  		float temp;
  		int pressure;
} imu6410;

imu6410 volatile imu6410_raw;

void PrintAccelGyroMag()
{
 	 printf("%.3f %.3f %.3f  ",imu6410_raw.xfaccel,imu6410_raw.yfaccel,
	                           imu6410_raw.zfaccel);
 	 printf("%.3f %.3f %.3f  ",imu6410_raw.xfgyro,imu6410_raw.yfgyro,
	                           imu6410_raw.zfgyro);
 	 printf("%.3f %.3f %.3f  ",imu6410_raw.xfmag,imu6410_raw.yfmag,
	                           imu6410_raw.zfmag);
}

void ReadAccelGyroMag()
{
	imu6410_raw.xaccel = SPI_ADXL345_ReadAxis(ADXL345XAXIS);
	imu6410_raw.yaccel = SPI_ADXL345_ReadAxis(ADXL345YAXIS);
	imu6410_raw.zaccel = SPI_ADXL345_ReadAxis(ADXL345ZAXIS);
	imu6410_raw.xgyro = SPI_L3G4200D_ReadAxis(L3G4200XAXIS);
	imu6410_raw.ygyro = SPI_L3G4200D_ReadAxis(L3G4200YAXIS);
	imu6410_raw.zgyro = SPI_L3G4200D_ReadAxis(L3G4200ZAXIS);
	imu6410_raw.xmag = HMC5883_ReadAxis(HMC5883_XAXIS);
	imu6410_raw.ymag = HMC5883_ReadAxis(HMC5883_YAXIS);
	imu6410_raw.zmag = HMC5883_ReadAxis(HMC5883_ZAXIS);
}

void ConvertIMU6410Raw()
{
	imu6410_raw.xfaccel = (float)imu6410_raw.xaccel * 0.0039;	
	imu6410_raw.yfaccel = (float)imu6410_raw.yaccel * 0.0039;
	imu6410_raw.zfaccel = (float)imu6410_raw.zaccel * 0.0039;
	imu6410_raw.xfgyro = ((float) imu6410_raw.xgyro)*250.0/32768.0;
	imu6410_raw.yfgyro = ((float) imu6410_raw.ygyro)*250.0/32768.0;
	imu6410_raw.zfgyro = ((float) imu6410_raw.zgyro)*250.0/32768.0;
	imu6410_raw.xfmag = ((float) imu6410_raw.xmag)*1.3/2048.0;
	imu6410_raw.yfmag = ((float) imu6410_raw.ymag)*1.3/2048.0;
	imu6410_raw.zfmag = ((float) imu6410_raw.zmag)*1.3/2048.0;
}

void CorrectAccelGyro()
{
// Unbias Gyro Data
	imu6410_raw.xfgyro -= GyroBias[0];
	imu6410_raw.yfgyro -= GyroBias[1];
	imu6410_raw.zfgyro -= GyroBias[2];
	
// Unbias Accel Data
   imu6410_raw.xfaccel -= AccelBias[0];
   imu6410_raw.yfaccel -= AccelBias[1];
   imu6410_raw.zfaccel += 1.00 - AccelBias[2];
}

void CorrectGyroMagnetometer()
{
	float temp=0.0,magmag=0.0;

	CorrectAccelGyro();
	
// Scale Magnetometer Values (No Need)
//	magmag = pow(fabs(imu6410_raw.xfmag),2) + pow(fabs(imu6410_raw.yfmag),2)
//			       + pow(fabs(imu6410_raw.zfmag),2);
	magmag = imu6410_raw.xfmag*imu6410_raw.xfmag + imu6410_raw.yfmag*imu6410_raw.yfmag
			       + imu6410_raw.zfmag*imu6410_raw.zfmag;
	printf(" magmag=%8.3f ",magmag);
//	magmag=pow(abs(pkt_ptr[10]),2)+pow(abs(pkt_ptr[11]),2)+pow(abs(pkt_ptr[12]),2);
	imu6410_raw.xfmag /= sqrt(magmag);
	imu6410_raw.yfmag /= sqrt(magmag);
	imu6410_raw.zfmag /= sqrt(magmag);

// Change Axes to Align with North East Down (NED) frame
	// new ax    =  - old ay
	// new ay    =  - old ax
	// new az    =  - old az
	temp	     = imu6410_raw.xfaccel;
	imu6410_raw.xfaccel  = -imu6410_raw.yfaccel;
	imu6410_raw.yfaccel  = -temp;
	imu6410_raw.zfaccel *= -1.0;

	// new gx    =  - old gx
	// new gy    =  - old gy
	// gz stays the same
	imu6410_raw.xfgyro *=  -1.0;
	imu6410_raw.yfgyro *=  -1.0;

	// new mx    =  + old my
	// new my    =  + old mx
	// new mz    =  - old mz
	temp	     = imu6410_raw.xfmag;
	imu6410_raw.xfmag = imu6410_raw.yfmag;
	imu6410_raw.yfmag = temp;
	imu6410_raw.zfmag *= -1.0;
}

char DisplayDOF10WValues(unsigned char inchar)
{
    char inchar2;
	int delay=40;
	
    while((inchar2=getcharnw())!='e') {
		if(inchar2=='s') {
		    printf("Enter speed (msec): ");
			delay = readint();
			}
	    if(inchar=='a')  {
		    DisplayADXL345Data();
			putchar('\r');
			}
		if(inchar=='g')  {
		    DisplayGyroL3G4200();
			putchar('\r');
			}
		if(inchar=='m')  {
		    DisplayMagnetometer();
			putchar('\r');
			}
		if(inchar=='p')  {
		    DisplayMagnetometerPlusMinus();
			return 'p';
			}
		if(inchar=='b')  {
		    DisplayBMP085Brief('c');
			putchar('\r');
			}
		if(inchar=='1')  {
		    printf("\n\r");
		    DisplayDOF10WData('c');
			putchar('\r');
			}
	    Delaymsec(delay);
		}
	return 'c';
}

void DisplayDOF10WRegisters(unsigned char inchar)
{	
	if(inchar=='a') ;
    if(inchar=='g')  DumpL3G4200Registers();
	if(inchar=='m')  DumpHMC5883Regs();
	if(inchar=='b')  ;
}

#ifdef IMU6410
const char dof10mess1[] PROGMEM = "\n\rIMU6410 Menu\n\r";
#else 
const char dof10mess1[] PROGMEM = "\n\rIMU6420 Menu\n\r";

#endif

const char dof10mess2[] PROGMEM = "da - Display Acceleration\n\r";
const char dof10mess3[] PROGMEM = "dg - Display Gyro\n\r";
const char dof10mess4[] PROGMEM = "dm - Display Magnetometer\n\r";
const char dof10mess45[] PROGMEM = "dp - Display Magnetometer Plus-Negative bias\n\r";
const char dof10mess5[] PROGMEM = "db - Display Barometer\n\r";
const char dof10mess6[] PROGMEM = "d1 - Display All\n\r";
const char dof10mess7[] PROGMEM = "ra - Dump Acceleration Registers\n\r";
const char dof10mess8[] PROGMEM = "rg - Display Gyro Registers\n\r";
const char dof10mess9[] PROGMEM = "rm - Display Magnetometer Registers\n\r";
const char dof10mess10[] PROGMEM = "rb - Display Barometer Registers\n\r";
const char dof10mess11[] PROGMEM = "\n\rDOF-";

void DOF10WMenu(void)
{
 	char inchar1,inchar2;
	
    putfs(dof10mess1);
    putfs(dof10mess2);
    putfs(dof10mess3);
    putfs(dof10mess4);
	putfs(dof10mess45);
    putfs(dof10mess5);
    putfs(dof10mess6);
    putfs(dof10mess7);
    putfs(dof10mess8);
    putfs(dof10mess9);
	putfs(dof10mess11);
	while((inchar1=getchar())!='e') {
	    putchar(inchar1);
	    inchar2 = getchar();
		putchar(inchar2);
		printf("\n\r");
		if(inchar1=='d') DisplayDOF10WValues(inchar2);
		if(inchar1=='r') DisplayDOF10WRegisters(inchar2);
	    putfs(dof10mess11);
	    }
}

void Accel_Calib()
{
		int i=0, N_samples=50;
		float sumx=0.0,sumy=0.0,sumz=0.0;

		printf("Calculating Accelorometer Bias \n\r");

		for(i=0; i<N_samples; i++){
			ReadAccelGyroMag();
			ConvertIMU6410Raw();
			
			sumx += imu6410_raw.xfaccel;
			sumy += imu6410_raw.yfaccel;
			sumz += imu6410_raw.zfaccel;
			
			Delaymsec(20);
		    }
		AccelBias[0] = sumx / N_samples;
		AccelBias[1] = sumy / N_samples;
		AccelBias[2] = sumz / N_samples;
		
		printf("Accel Bias x y z= %.3f %.3f %.3f \n\r",
		                     AccelBias[0],AccelBias[1],AccelBias[2]);
}

void Gyro_Calib()
{
		int i=0, N_samples=50;
		float sumx=0.0,sumy=0.0,sumz=0.0;

		float temp_zero[3]; // this is a dummy just to call Getpkt

		printf("\n\rCalculating Gyro Bias \n\r");

		for (i=0;i<3;i++){
			temp_zero[i]=0.0;	
		    }

		for(i=0; i<N_samples; i++){
			//WASP_Get_Packet(&data_pkt[0],&temp_zero[0]);
			ReadAccelGyroMag();
			ConvertIMU6410Raw();
			
			sumx += imu6410_raw.xfgyro;
			sumy += imu6410_raw.yfgyro;
			sumz += imu6410_raw.zfgyro;
			
			Delaymsec(20);
		    }
		GyroBias[0] = sumx / N_samples;
		GyroBias[1] = sumy / N_samples;
		GyroBias[2] = sumz / N_samples;
		
		printf("Gyro Bias x y z= %.3f %.3f %.3f \n\r",
		                     GyroBias[0],GyroBias[1],GyroBias[2]);
}

float wrap(float angle){   // I/P and O/P in degrees
	
	angle *= (PI/180);  // deg to rad

	if (angle > PI){
		angle-= (2*PI);
	}
	else if (angle < -PI){
		angle+= (2*PI);
	}
	else if (angle ==0){
		angle+= (2*PI);		
	}

	return (angle*(180/PI));   // return angle in deg
}

#define COMPLEMENTARYFILERRATE 10
//#define COMPLEMENTARYFILERRATE 1000

// This filter is a work in progress and is not producing the correct 
// values at this time.
void  ComplementaryFilter()
{
	float delta_t_sec=0.0,accumulatedtime=0.0;
	
	float gyro_roll=0.0,gyro_pitch=0.0,gyro_yaw=0.0;
	float acc_roll=0.0,acc_pitch=0.0;
	
	float a=0.0;
	float comp_roll=0.0,comp_pitch=0.0,comp_yaw=0.0;
	float comp_roll_rad=0.0,comp_pitch_rad=0.0,comp_yaw_rad=0.0;
	
	float xh=0.0,xh1,xh2, yh=0.0;

 	printf("\n\rSOC Complimentary Roll,Pitch,Yaw Filter\n\r");
	
	Gyro_Calib();
	
	// Set filter cycle rate   
	mseccountdown4 = COMPLEMENTARYFILERRATE;
	
	while(!getcharnw()){	// main loop
		
		while(mseccountdown4) ;
	    mseccountdown4 = COMPLEMENTARYFILERRATE;
		
		//WASP_Get_Packet(&data_pkt[0],&Gyrobias[0]);
		ReadAccelGyroMag();
		ConvertIMU6410Raw();
		CorrectGyroMagnetometer();

		//	Time: hh=data_pkt[0] min=data_pkt[1] sec=data_pkt[2] millisec=data_pkt[3]
		//  Acc:  ax=data_pkt[4] ay=data_pkt[5] az=data_pkt[6]
		//  Gyr:  gx=data_pkt[7] gy=data_pkt[8] gz=data_pkt[9]
		//  Mag:  mx=data_pkt[10] my=data_pkt[11] mz=data_pkt[12]

		// To Be fixed
		//  Temperature: T=data_pkt[13]
		//  Pressure:    P=data_pkt[14]

// Print Sensor Packet
		//t2=((data_pkt[0]*60*60*1000)+(data_pkt[1]*60*1000)+(data_pkt[2]*1000)+data_pkt[3]);
		//delta_t_ms=t2-t1;
		//t1=t2;		
		//delta_t_sec=delta_t_ms/1000;
		delta_t_sec = ((float) COMPLEMENTARYFILERRATE)/1000.0;
		accumulatedtime += delta_t_sec;

// Orientation only from Gyros  <<<<<<<< SOMETHING IS WRONG Gives Huge start Bias 600s-700s (is it start up time) ?

		gyro_roll += imu6410_raw.xfgyro * delta_t_sec;
		gyro_pitch += imu6410_raw.yfgyro * delta_t_sec;
		gyro_yaw += imu6410_raw.zfgyro * delta_t_sec;

//	printf("Gyro_Roll=%lf\tGyro_Pitch=%lf\tGyro_Yaw=%lf\t",gyro_roll,gyro_pitch,gyro_yaw);
		
// Roll and Pitch from Accelerometers only

		// Roll = atan2 (ay/az)
		acc_roll = -(atan2(imu6410_raw.yfaccel,-imu6410_raw.zfaccel))*(180/PI);

		// Pitch = atan2 (ax/az)
		acc_pitch = (atan2(imu6410_raw.xfaccel,-imu6410_raw.yfaccel))*(180/PI);

//		printf("AccRoll= %lf\tAccPitch= %lf\t",acc_roll,acc_pitch);
		
// Complimentary Filter (Roll and Pitch)  LPF Acc , HPF Gyros
		a = 0.98;
		comp_roll = (a  * (comp_roll+(imu6410_raw.yfgyro*delta_t_sec)) )+ ((1-a) * (acc_roll));
		comp_pitch = (a * (comp_pitch+(imu6410_raw.yfgyro*delta_t_sec)) )+ ((1-a) * (acc_pitch));
		
		//printf("CompRoll= %lf\tCompPitch= %lf\t",comp_roll,comp_pitch);
		
// Magnetometer Tilt Compensation

		comp_pitch_rad = comp_pitch *(PI/180);
		comp_roll_rad = comp_roll *(PI/180);

// Magnetometer Tilt Compensation Works  (Just Fine TUNE , don't alter)
		xh = (imu6410_raw.xfmag*cos(comp_pitch_rad)); 
		xh1 = (imu6410_raw.yfmag*sin(comp_pitch_rad)*sin(comp_roll_rad));
		xh2 = (imu6410_raw.zfmag*cos(comp_roll_rad)*sin(comp_pitch_rad)); 
		xh += xh1 + xh2;
		
		yh = imu6410_raw.yfmag*cos(comp_roll_rad)-
			 imu6410_raw.zfmag*sin(comp_roll_rad);
		
		comp_yaw_rad = wrap(atan2(yh,xh));  // not sure if it is yh or -yh
		comp_yaw = comp_yaw_rad *(180/PI);

		printf("%7.3f: ",accumulatedtime);
		if(0) PrintAccelGyroMag();

		printf("%8.3f %8.3f %8.3f  \r",comp_roll,comp_pitch,comp_yaw); 
		}
}

void CompassHeading()
{
	printf("\n\rCompass Heading:\n\r");

}
// TBD
void KalmanFilter()
{
    printf("\n\rKalman Filter not implemented");
}

#define DEADRECKONRATE  30

// Simple dead recking function - drifts very quickly
void DeadReckon()
{
 	char inchar;
  	float xpos,ypos,zpos;
	float delta_t_sec=0.0,delta_t_sqr,accumulatedtime=0.0;
	float gyro_roll=0.0,gyro_pitch=0.0,gyro_yaw=0.0;
	
    printf("\n\rDead Reckon Output\n\r");
 	Gyro_Calib();
	Accel_Calib();
	printf("Commands: r-reset, p-pause, e-exit\n\r");
	xpos = ypos = zpos = gyro_roll = gyro_pitch = gyro_yaw = 0.0;
	
	mseccountdown4 = DEADRECKONRATE;
	
	while((inchar=getcharnw())!='e'){
	
		if(inchar=='r') {
		    xpos = ypos = zpos = 0.0;
			gyro_roll = gyro_pitch = gyro_yaw = 0;
			}
			
		while(mseccountdown4) ;
	    mseccountdown4 = DEADRECKONRATE;

		ReadAccelGyroMag();
		ConvertIMU6410Raw();
		CorrectAccelGyro();
		delta_t_sec = ((float) DEADRECKONRATE)/1000.0;
				
		if(inchar!='p') {  
		    accumulatedtime += delta_t_sec;
		    delta_t_sqr = delta_t_sec * delta_t_sec;
		
		    xpos += 39.37 * 9.807 * imu6410_raw.xfaccel * delta_t_sqr;
		    ypos += 39.37 * 9.807 * imu6410_raw.yfaccel * delta_t_sqr;
		    zpos -= 39.37 * 9.807 * (1.00 - imu6410_raw.zfaccel) * delta_t_sqr;
		
		    gyro_roll += imu6410_raw.xfgyro * delta_t_sec;
		    gyro_pitch += imu6410_raw.yfgyro * delta_t_sec;
		    gyro_yaw += imu6410_raw.zfgyro * delta_t_sec;
		
		    if(1) printf("%#6.3f %#6.3f %#6.3f  %#6.3f %#6.3f %#6.3f \r",
			            xpos,ypos,zpos,
						imu6410_raw.xfaccel,imu6410_raw.yfaccel,imu6410_raw.zfaccel);
		    if(0) printf("%#6.3f %#6.3f %#6.3f  %#6.3f %#6.3f %#6.3f   \n\r",
			            imu6410_raw.xfaccel,imu6410_raw.yfaccel,imu6410_raw.zfaccel,
					 	imu6410_raw.xfgyro,imu6410_raw.yfgyro,imu6410_raw.zfgyro);
		    if(0) printf("%#6.3f %#6.3f %#6.3f  %#6.3f %#6.3f %#6.3f   \r",xpos,ypos,zpos,
					 	  			 	  gyro_roll,gyro_pitch,gyro_yaw);
			}
	    }
}

const char IMU6410mess1[] PROGMEM = "\n\rIMU6410 Help Menu\n\r";
const char IMU6410mess2[] PROGMEM = "  i - DOF10 menu\n\r";
const char IMU6410mess3[] PROGMEM = "  c - Complementary Filter - beta\n\r";
const char IMU6410mess4[] PROGMEM = "  f - Kalman Filter - TBD\n\r";
const char IMU6410mess5[] PROGMEM = "  d - Display 9DOF data\n\r";
const char IMU6410mess6[] PROGMEM = "  g - Display calibrated 9DOF data\n\r";
const char IMU6410mess7[] PROGMEM = "  b - Calibrate gyro and accelerometer\n\r";
const char IMU6410mess8[] PROGMEM = "  p - Start Dead reckoning\n\r";
const char IMU6410mess9[] PROGMEM = "  ? - Help menu\n\r";

void IMU6410Help()
{
    putfs(IMU6410mess1);
    putfs(IMU6410mess2);
    putfs(IMU6410mess3);
    putfs(IMU6410mess4);
    putfs(IMU6410mess5);
    putfs(IMU6410mess6);
    putfs(IMU6410mess7);
    putfs(IMU6410mess8);
    putfs(IMU6410mess9);
}

// Main IMU6410 menu
void IMU6410Menu()
{
 	 char inchar,inchar2;
	 
 	 printf("\n\rIMU6410 Main Menu\n\r");
	 printf("Enter command [i,c,f,d,g,b,p,?] or 'e' to exit\n\rIMU6410-");
	 inchar = ' ';
	 while((inchar=getchar())!='e') {
	 	putchar(inchar);
		switch (inchar) {
			case 'i':
			    DOF10WMenu();
				break;
			case 'c':
				ComplementaryFilter();
				break;
			case 'f':
				KalmanFilter();
				break;
			case 'd':
				putchar(inchar2=getchar());
				printf("\n\r");
				DisplayDOF10WData(inchar2);
				printf("\n\r");
				ReadAccelGyroMag();
				ConvertIMU6410Raw();
				CorrectAccelGyro();
				PrintAccelGyroMag();
				break;
			case 'g':
				putchar(inchar2=getchar());
				Gyro_Calib(&GyroBias[0]);
				Accel_Calib(&AccelBias[0]);
				DisplayDOF10WData(inchar2);
				printf("\n\r");
				ReadAccelGyroMag();
				ConvertIMU6410Raw();
				CorrectAccelGyro();
				PrintAccelGyroMag();
				break;
			case 'b':
				Gyro_Calib();
				Accel_Calib();
				break;
			case 'p':
				DeadReckon();
				break;
			case '?':
				IMU6410Help();
				break;
		    }
			printf("\n\rIMU6410-");
	    }
}

